package ai.lazuli.cesga;

import java.io.File;
import java.util.Arrays;

import ai.lazuli.cesga.data.ModelBuilder;
import ai.lazuli.cesga.data.ModelSynchronizer;
import ai.lazuli.cesga.data.Purger;
import ai.lazuli.cesga.data.collection.BuildModelAfterCollection;
import ai.lazuli.cesga.data.collection.Collector;
import ai.lazuli.cesga.data.collection.PredictAfterCollection;
import ai.lazuli.cesga.logging.Logger;

public final class StartAI {
	// *** EXEC *** //
	// ************ //
	public static void exec(String cfgPath) {
		// Obtain config
		Config cfg = Config.getInstance(cfgPath);
		
		// Configure logger
		Logger.LOGS_DIR = cfg.getLogsPath();
		Logger.println("STARTING AI for jobs classification ...");
		
		// Build necessary directories
		buildDirs(cfg);
		
		// Start purger
		startShortDataPurger(cfg);
		if(cfg.isLongDataEnabled())
			startLongDataPurger(cfg);
		startPredictorPredictionsPurger(cfg);
		startPredictorCollectionsPurger(cfg);
		startLogsPurger(cfg);
		
		// Start collectors
		ModelSynchronizer shortSync = new ModelSynchronizer();
		ModelSynchronizer longSync = new ModelSynchronizer();
		startShortCollector(cfg, shortSync);
		if(cfg.isLongDataEnabled())
			startLongCollector(cfg, longSync);
		startPredictor(cfg, shortSync, longSync);
		
		Logger.println("AI for jobs classification STARTED");
	}
	
	// *** AUX FUNCTIONS *** //
	// ********************* //
	public static void buildDirs(Config cfg) {
		buildDir(cfg.getLogsPath());
		buildDir(cfg.getLongDataPath());
		buildDir(cfg.getShortDataPath());
		buildDir(cfg.getPredictorPath());
	}
	
	public static void buildDir(String path) {
		File dir = new File(path);
		dir.mkdirs();
		if(!dir.exists()) {
			String err = String.format("Could NOT make necessary directories.\n"+
					"\tFailed to make: '%s'\n", dir.getAbsolutePath());
			System.err.print(err); // Because logs directory may not exist
			Logger.print(err);
		}
	}
	
	public static Purger startShortDataPurger(Config cfg) {
		long ttl = cfg.getShortDataTimeToLive();
		if(ttl <= 0) return null;
		
		Purger p = new Purger(ttl/3,ttl,
				new File(cfg.getShortDataPath()));
		p.startPurging();
		return p;
	}
	
	public static Purger startLongDataPurger(Config cfg) {
		long ttl = cfg.getLongDataTimeToLive();
		if(ttl <= 0) return null;
		
		Purger p = new Purger(ttl/3, ttl,
				new File(cfg.getLongDataPath()));
		p.startPurging();
		return p;
	}
	
	public static Purger startLogsPurger(Config cfg) {
		long ttl = cfg.getLogsTimeToLive();
		if(ttl <= 0) return null;
		
		Purger p = new Purger(ttl/3, ttl,
				new File(cfg.getLogsPath()));
		p.startPurging();
		return p;
	}
	
	public static Purger startPredictorPredictionsPurger(Config cfg) {
		long ttl = cfg.getPredictorPredictionsTimeToLive();
		if(ttl <= 0) return null;
		
		Purger p = new Purger(ttl/3, ttl,
				new File(cfg.getPredictorPredictionsPath())
				);
		p.startPurging();
		return p;
	}
	
	public static Purger startPredictorCollectionsPurger(Config cfg) {
		long ttl = cfg.getPredictorCollectionsTimeToLive();
		if(ttl <= 0) return null;
		
		Purger p = new Purger(ttl/3, ttl,
				new File(cfg.getPredictorCollectionPath())
				);
		p.startPurging();
		return p;
	}
	
	public static Collector startShortCollector(Config cfg, ModelSynchronizer shortSync) {
		try {
			Collector c = new Collector(
					cfg.getShortCollectionInterval(),
					true,
					true,
					cfg.getShortCollectionScript(),
					cfg.getShortDataPath(),
					new BuildModelAfterCollection(
						cfg.getShortModelPath(),
						new ModelBuilder(
							cfg.getShortModelScript(),
							shortSync
						)
					));
			c.start();
			return c;
		}catch(Exception ex) {
			String err = "ERROR starting short collector";
			Logger.println(err);
			Logger.printException(ex);
			System.exit(1);
		}
		return null; // Should NEVER be reached because of System.exit inside catch
	}
	
	public static Collector startLongCollector(Config cfg, ModelSynchronizer longSync) {
		try {
			Collector c = new Collector(
					cfg.getLongCollectionInterval(),
					true,
					true,
					cfg.getLongCollectionScript(),
					cfg.getLongDataPath(),
					new BuildModelAfterCollection(
						cfg.getLongModelPath(),
						new ModelBuilder(
							cfg.getLongModelScript(),
							longSync
						)
					));
			c.start();
			return c;
		}catch(Exception ex) {
			String err = "ERROR starting long collector";
			Logger.println(err);
			Logger.printException(ex);
			System.exit(1);
		}
		return null; // Should NEVER be reached because of System.exit inside catch
	}
	
	public static Collector startPredictor(Config cfg, ModelSynchronizer shortSync, ModelSynchronizer longSync) {
		try {
			Collector p = new Collector(cfg.getPredictorInterval(),
					false,
					true,
					cfg.getPredictorScriptPath(),
					cfg.getPredictorCollectionPath(),
					new PredictAfterCollection(
							cfg.getShortModelPath(),
							cfg.getLongModelPath(),
							shortSync,
							longSync,
							cfg.getPredictorPredictionsPath(),
							null,
							cfg)
					);
			((PredictAfterCollection)p.getAfterCollectionStrategy()).setPredictor(p);
			String oldCmd[] = p.getCollectionCommand();
			String newCmd[] = Arrays.copyOf(oldCmd, oldCmd.length+3);
			newCmd[newCmd.length-3] = cfg.getPredictorLastCollectionEpoch()+""; // $2
			newCmd[newCmd.length-2] = (System.currentTimeMillis()/1000)+""; // $3
			newCmd[newCmd.length-1] = cfg.getNotifierPool(); // $4
			cfg.setPredictorLastCollectionEpoch(Long.parseLong(newCmd[newCmd.length-2]));
			cfg.save();
			p.setCollectionCommand(newCmd);
			p.start();
			return p;
		}catch(Exception ex) {
			String err = "ERROR starting predictor";
			Logger.println(err);
			Logger.printException(ex);
			System.exit(1);
		}
		return null; // Should NEVER be reached because of System.exit inside catch
	}
}
