package ai.lazuli.cesga.data;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;

import ai.lazuli.cesga.logging.Logger;

public class ModelBuilder {
	// *** CONSTANTS *** //
	// ***************** //
	public final static int R_SCRIPT_EXPECTED_EXIT_CODE = 0; 
	
	// *** ATTRIBUTES *** //
	// ****************** //
	private String scriptPath; // Path to the R script used to build the model
	private ModelSynchronizer synchronizer; // Used to handle synchronized access to shared resources (model files)
	
	// *** CONSTRUCTORS *** //
	// ******************** //
	public ModelBuilder(String scriptPath, ModelSynchronizer synchronizer) {
		setScriptPath(scriptPath);
		setModelSynchronizer(synchronizer);
	}
	
	// *** GETTERs and SETTERs *** //
	// *************************** //
	public String getScriptPath() {
		return scriptPath;
	}

	public void setScriptPath(String scriptPath) {
		this.scriptPath = scriptPath;
	}
	
	public ModelSynchronizer getModelSynchronizer() {
		return synchronizer;
	}

	public void setModelSynchronizer(ModelSynchronizer synchronizer) {
		this.synchronizer = synchronizer;
	}

	// *** METHODS *** //
	// *************** //
	/**
	 * Generates a model
	 * @param csvPath Path to the CSV file which contains training data.
	 * @param modelPath Path where the model rules will be exported.
	 */
	public void build(String csvPath, String modelPath) {
		String cwd = new File("").getAbsolutePath()+File.separator;
		String cmd[] = new String[]{"Rscript",
				String.format("%s%s", cwd, getScriptPath()),
				String.format("%s%s", cwd, csvPath),
				String.format("%s%s", cwd, modelPath)};
		try {
			lockWriteMutex();
			Process p = Runtime.getRuntime().exec(cmd);
			int code = p.waitFor();
			if(code == R_SCRIPT_EXPECTED_EXIT_CODE) {
				Logger.printf("Model was successfully built at \"%s\" with data from \"%s\"\n",
						modelPath, csvPath);
			}else {
				BufferedReader in = new BufferedReader(new InputStreamReader(p.getInputStream()));
				String output = "", read;
				while( (read=in.readLine()) != null) 
					output += read+"\n";
				in.close();
				in = new BufferedReader(new InputStreamReader(p.getErrorStream()));
				while( (read=in.readLine()) != null) 
					output += read+"\n";
				in.close();
				
				Logger.printf("FAILED to build model at \"%s\" with data from \"%s\"\n"+
						"Used command: '\"%s %s %s %s' \nOutput: %s",
						modelPath, csvPath, cmd[0], cmd[1], cmd[2], cmd[3], output);
			}
		}catch(Exception ex) {
			Logger.printf("Error building model \"%s\" with data from \"%s\"\nUsed command was: "
					+"'%s %s %s %s'\n",
					modelPath, csvPath, cmd[0], cmd[1], cmd[2], cmd[2], cmd[3]);
			Logger.printException(ex);
		}
		finally {
			releaseWriteMutex();
		}
	}
	
	// *** INNER SYNC METHODS *** //
	// ************************** //
	private void lockWriteMutex() {
		while(!getModelSynchronizer().lockWriteMutex()) {
			try {
				Thread.sleep(5000);
			}catch(InterruptedException iex) {
				Logger.printException(iex);
			}
		};
	}
	private void releaseWriteMutex() {
		getModelSynchronizer().unlockWriteMutex();
	}

}
