package ai.lazuli.cesga.data;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.PrintWriter;
import java.util.List;
import java.util.Properties;

import ai.lazuli.cesga.Config;
import ai.lazuli.cesga.Queries;
import ai.lazuli.cesga.data.collection.PredictAfterCollection;
import ai.lazuli.cesga.logging.Logger;
import ai.lazuli.utils.csv.CSV;
import ai.lazuli.utils.csv.CSVIO;

public class Notifier {
	// *** CONSTANTS *** //
	// ***************** //
	public final static String RUNNING_STATE = "RUNNING";
	// *** CFG FIELD NAMES *** //
	public final static String MIN_ACC = "MinAccuracy"; // Min accuracy criteria for considering a prediction
	public final static String STATE_FILTER = "StateFilter";
	public final static String USER_FILTER = "UserFilter";
	public final static String ACCOUNT_FILTER = "AccountFilter";
	public final static String MIN_MEM_REQ = "MinMemReq";
	public final static String MAX_MEM_REQ = "MaxMemReq";
	public final static String MIN_CPUS_REQ = "MinCpusReq";
	public final static String MAX_CPUS_REQ = "MaxCpusReq";
	public final static String MIN_CPUS_ALLOC = "MinCpusAlloc";
	public final static String MAX_CPUS_ALLOC = "MaxCpusAlloc";
	public final static String MIN_START_TIME = "MinStartTime";
	public final static String MAX_START_TIME = "MaxStartTime";
	public final static String MIN_NODES_ALLOC = "MinNodesAlloc";
	public final static String MAX_NODES_ALLOC = "MaxNodesAlloc";
	public final static String MIN_ELAPSED_TIME = "MinElapsedTime";
	public final static String MAX_ELAPSED_TIME = "MaxElapsedTime";
	public final static String START_WEEK_DAY_FILTER = "StartWeekDayFilter";
	public final static String PARTITION_FILTER = "PartitionFilter";
	public final static String NODE_FILTER = "NodeFilter"; // Nodelist field related
	
	// *** ATTRIBUTES *** //
	// ****************** //
	private String notificationsPath; // Path to file where notifications are stored
	private String poolPath; // Path to the file which contains the list of job ids which must be queried
	private Properties cfg; // Used to store notifier configuration in persistent storage
	
	// *** SINGLETON *** //
	// ***************** //
	private static Notifier notifier = null;
	
	private Notifier() {}
	
	public static Notifier getInstance(Config config) {
		if(notifier==null)
			notifier = new Notifier();
		notifier.setNotificationsPath(config.getNotificationsPath());
		notifier.setPoolPath(config.getNotifierPool());
		notifier.generateCfgIfNotExists(config);
		return notifier;
	}
	
	private void generateCfgIfNotExists(Config config)  {
		String path = config.getNotifierConfiguration();
		setCfg(new Properties());
		if(new File(path).exists()) {
			try {
				getCfg().load(new FileReader(path));
				return;
			}catch(Exception ex) {
				Logger.printException(ex);
			}
		}
		
		getCfg().setProperty(MIN_ACC, ""+0.95);
		getCfg().setProperty(STATE_FILTER, "[FAILED,TIMEOUT]");
		getCfg().setProperty(USER_FILTER, "null");
		getCfg().setProperty(ACCOUNT_FILTER, "[other]");
		getCfg().setProperty(MIN_MEM_REQ, "null");
		getCfg().setProperty(MAX_MEM_REQ, "null");
		getCfg().setProperty(MIN_CPUS_REQ, "null");
		getCfg().setProperty(MAX_CPUS_REQ, "null");
		getCfg().setProperty(MIN_CPUS_ALLOC, "null");
		getCfg().setProperty(MAX_CPUS_ALLOC, "null");
		getCfg().setProperty(MIN_START_TIME, "null");
		getCfg().setProperty(MAX_START_TIME, "null");
		getCfg().setProperty(MIN_NODES_ALLOC, "null");
		getCfg().setProperty(MAX_NODES_ALLOC, "null");
		getCfg().setProperty(MIN_ELAPSED_TIME, "1800");
		getCfg().setProperty(MAX_ELAPSED_TIME, "null");
		getCfg().setProperty(START_WEEK_DAY_FILTER, "null");
		getCfg().setProperty(PARTITION_FILTER, "null");
		getCfg().setProperty(NODE_FILTER, "null");
		
		try {
			getCfg().store(new PrintWriter(path), null);
		}catch(Exception ex) {
			Logger.printException(ex);
		}
	}
	
	// *** NOTIFY *** //
	// ************** //
	public void notify(CSV input){
		// Apply notifier filters/queries
		File parent = new File(getNotificationsPath()).getParentFile();
		if(parent!=null && !parent.exists() && !parent.mkdirs()) {
			Logger.printf("Error creating directory \"%s\"\n"+
					"No notifications will be exported.\n",
					parent.getAbsolutePath());
			return;
		}
		
		CSV output = input;
		try {
			if(getStateFilter() != null)
				output = Querier.getJobsByPredictedState(output, getStateFilter());
			if(getUserFilter() != null)	
				output = Querier.getJobsByUser(output, getUserFilter());
			if(getAccountFilter() != null)
				output = Querier.getJobsByAccount(output, getAccountFilter());
			if(getStartWeekDayFilter() != null)
				output = Querier.getJobsByStartWeekDay(output, getStartWeekDayFilter());
			if(getPartitionFilter() != null)
				output = Querier.getJobsByPartition(output, getPartitionFilter());
			if(getNodeFilter() != null)
				output = Querier.getJobsByNodelist(output, getNodeFilter());
			output = Querier.getJobsByMemReq(output, getMinMemReq(), getMaxCpusReq());
			output = Querier.getJobsByCpusReq(output, getMinCpusReq(), getMaxCpusReq());
			output = Querier.getJobsByCpusAlloc(output, getMinCpusAlloc(), getMaxCpusAlloc());
			output = Querier.getJobsByStartTime(output, getMinStartTime(), getMaxStartTime());
			output = Querier.getJobsByNodesAlloc(output, getMinNodesAlloc(), getMaxNodesAlloc());
			output = Querier.getJobsByElapsedTime(output, getMinElapsedTime(), getMaxElapsedTime());
		}catch(QuerierException qex) {
			Logger.println("Could not filter notification");
			Logger.printException(qex);
			return;
		}
		
		// Export/update pool
		try {
			PrintWriter out = new PrintWriter(getPoolPath());
			for(int i = 0 ; i < output.size() ; i++) {
				/*
				 *  It is not necessary to check state is running because
				 *  the script used to collect by predictor only retrieves
				 *  running jobs.
				 *  
				 *  If that changed, just change the for-loop body by what it
				 *  is commented here.
				 */
				/*String state = output.getValue(i, Querier.FIELD_STATE).toString();
				if(state.equals(RUNNING_STATE)) {
					out.println(output.getValue(i, PredictAfterCollection.FIELD_ID_JOB).toString());
				}*/
				out.println(output.getValue(i, PredictAfterCollection.FIELD_ID_JOB).toString());
			}
			out.close();
		}catch(FileNotFoundException fnfex) {
			Logger.printf("Notifier could not write to the pool at \"%s\"\n",
					getPoolPath());
			Logger.printException(fnfex);
		}
		
		// Export/update notifications
		try {
			if(!new File(getNotificationsPath()).exists()) {
				// First notifications export
				CSVIO.write(output, getNotificationsPath(), false);
			}
			else {	
				// After-first notifications export
				CSV notis = CSVIO.read(getNotificationsPath(), true, null);
				for(int i = 0 ; i < output.size() ; i++) {
					// Remove obsolete predictions
					String idOutput = output.getValue(i, PredictAfterCollection.FIELD_ID_JOB).toString();
					for(int j = 0 ; j < notis.size() ; j++) {
						String idNotis = notis.getValue(j, PredictAfterCollection.FIELD_ID_JOB).toString();
						if(idOutput.equals(idNotis)) {
							notis.delRecord(j);
							j--;
						}
					}
				}
				notis.addRecords(output.getRecords());
				CSVIO.write(notis, getNotificationsPath(), false);
			}
			Logger.printf("Notifications exported at \"%s\"\n", getNotificationsPath());
		}catch(Exception ex) {
			Logger.println("Could not export notification.");
			Logger.printException(ex);
		}
	}
	
	
	// *** GETTERs and SETTERs *** //
	// *************************** //
	public String getNotificationsPath() {
		return notificationsPath;
	}
	public void setNotificationsPath(String notificationsPath) {
		this.notificationsPath = notificationsPath;
	}

	public String getPoolPath() {
		return poolPath;
	}

	public void setPoolPath(String poolPath) {
		this.poolPath = poolPath;
	}

	public Properties getCfg() {
		return cfg;
	}
	public void setCfg(Properties cfg) {
		this.cfg = cfg;
	}
	
	public Double getMinAcc() {
		String s = cfg.getProperty(MIN_ACC);
		if(s.equals("null")) return null;
		return Double.parseDouble(s);
	}
	public void setMinAcc(double minAcc) {
		cfg.setProperty(MIN_ACC, minAcc+"");
	}
	
	public List<String> getStateFilter() {
		String s = cfg.getProperty(STATE_FILTER, "null");
		if(s.equals("null")) return null;
		return Queries.extractList(s);
	}
	public void setStateFilter(List<String> stateFilter) {
		setStateFilter(filterListToString(stateFilter));
	}
	public void setStateFilter(String stateFilter) {
		cfg.setProperty(STATE_FILTER, stateFilter);
	}
	
	public List<String> getUserFilter(){
		String s = cfg.getProperty(USER_FILTER, "null");
		if(s.equals("null")) return null;
		return Queries.extractList(s);
	}
	public void setUserFilter(List<String> userFilter) {
		setUserFilter(filterListToString(userFilter));
	}
	public void setUserFilter(String userFilter) {
		cfg.setProperty(USER_FILTER, userFilter);
	}
	
	public List<String> getAccountFilter(){
		String s = cfg.getProperty(ACCOUNT_FILTER, "null");
		if(s.equals("null")) return null;
		return Queries.extractList(s);
	}
	public void setAccountFilter(List<String> accountFilter) {
		setAccountFilter(filterListToString(accountFilter));
	}
	public void setAccountFilter(String accountFilter) {
		cfg.setProperty(ACCOUNT_FILTER, accountFilter);
	}
	
	public List<String> getStartWeekDayFilter(){
		String s = cfg.getProperty(START_WEEK_DAY_FILTER, "null");
		if(s.equals("null")) return null;
		return Queries.extractList(s);
	}
	public void setStartWeekDayFilter(List<String> startWeekDayFilter) {
		setStartWeekDayFilter(filterListToString(startWeekDayFilter));
	}
	public void setStartWeekDayFilter(String startWeekDayFilter) {
		cfg.setProperty(START_WEEK_DAY_FILTER, startWeekDayFilter);
	}
	
	public List<String> getPartitionFilter(){
		String s = cfg.getProperty(PARTITION_FILTER, "null");
		if(s.equals("null")) return null;
		return Queries.extractList(s);
	}
	public void setPartitionFilter(List<String> partitionFilter) {
		setPartitionFilter(filterListToString(partitionFilter));
	}
	public void setPartitionFilter(String partitionFilter) {
		cfg.setProperty(PARTITION_FILTER, partitionFilter);
	}
	
	public List<String> getNodeFilter(){
		String s = cfg.getProperty(NODE_FILTER, "null");
		if(s.equals("null")) return null;
		return Queries.extractList(s);
	}
	public void setNodeFilter(List<String> nodeFilter) {
		setNodeFilter(filterListToString(nodeFilter));
	}
	public void setNodeFilter(String nodeFilter) {
		cfg.setProperty(NODE_FILTER, nodeFilter);
	}
	
	public Long getMinMemReq() {
		String s = cfg.getProperty(MIN_MEM_REQ, "null");
		if(s.equals("null")) return null;
		return Long.parseLong(s);
	}
	public void setMinMemReq(Long minMemReq) {
		cfg.setProperty(MIN_MEM_REQ, ""+minMemReq);
	}
	
	public Long getMaxMemReq() {
		String s = cfg.getProperty(MAX_MEM_REQ, "null");
		if(s.equals("null")) return null;
		return Long.parseLong(s);
	}
	public void setMaxMemReq(Long maxMemReq) {
		cfg.setProperty(MAX_MEM_REQ, ""+maxMemReq);
	}
	
	public Long getMinCpusReq() {
		String s = cfg.getProperty(MIN_CPUS_REQ, "null");
		if(s.equals("null")) return null;
		return Long.parseLong(s);
	}
	public void setMinCpusReq(Long minCpusReq) {
		cfg.setProperty(MIN_CPUS_REQ, ""+minCpusReq);
	}
	
	public Long getMaxCpusReq() {
		String s = cfg.getProperty(MAX_CPUS_REQ, "null");
		if(s.equals("null")) return null;
		return Long.parseLong(s);
	}
	public void setMaxCpusReq(Long maxCpusReq) {
		cfg.setProperty(MAX_CPUS_REQ, ""+maxCpusReq);
	}
	
	public Long getMinCpusAlloc() {
		String s = cfg.getProperty(MIN_CPUS_ALLOC, "null");
		if(s.equals("null")) return null;
		return Long.parseLong(s);
	}
	public void setMinCpusAlloc(Long minCpusAlloc) {
		cfg.setProperty(MIN_CPUS_ALLOC, ""+minCpusAlloc);
	}
	
	public Long getMaxCpusAlloc() {
		String s = cfg.getProperty(MAX_CPUS_ALLOC, "null");
		if(s.equals("null")) return null;
		return Long.parseLong(s);
	}
	public void setMaxCpusAlloc(Long maxCpusAlloc) {
		cfg.setProperty(MAX_CPUS_ALLOC, ""+maxCpusAlloc);
	}
	
	public Long getMinStartTime() {
		String s = cfg.getProperty(MIN_START_TIME, "null");
		if(s.equals("null")) return null;
		return Long.parseLong(s);
	}
	public void setMinStartTime(Long minStartTime) {
		cfg.setProperty(MIN_START_TIME, ""+minStartTime);
	}
	
	public Long getMaxStartTime() {
		String s = cfg.getProperty(MAX_START_TIME, "null");
		if(s.equals("null")) return null;
		return Long.parseLong(s);
	}
	public void setMaxStartTime(Long maxStartTime) {
		cfg.setProperty(MAX_START_TIME, ""+maxStartTime);
	}
	
	public Long getMinNodesAlloc() {
		String s = cfg.getProperty(MIN_NODES_ALLOC, "null");
		if(s.equals("null")) return null;
		return Long.parseLong(s);
	}
	public void setMinNodesAlloc(Long minNodesAlloc) {
		cfg.setProperty(MIN_NODES_ALLOC, ""+minNodesAlloc);
	}
	
	public Long getMaxNodesAlloc() {
		String s = cfg.getProperty(MAX_NODES_ALLOC, "null");
		if(s.equals("null")) return null;
		return Long.parseLong(s);
	}
	public void setMaxNodesAlloc(Long maxNodesAlloc) {
		cfg.setProperty(MAX_NODES_ALLOC, ""+maxNodesAlloc);
	}
	
	public Long getMinElapsedTime() {
		String s = cfg.getProperty(MIN_ELAPSED_TIME, "null");
		if(s.equals("null")) return null;
		return Long.parseLong(s);
	}
	public void setMinElapsedTime(Long minElapsedTime) {
		cfg.setProperty(MIN_ELAPSED_TIME, ""+minElapsedTime);
	}
	
	public Long getMaxElapsedTime() {
		String s = cfg.getProperty(MAX_ELAPSED_TIME, "null");
		if(s.equals("null")) return null;
		return Long.parseLong(s);
	}
	public void setMaxElapsedTime(Long maxElapsedTime) {
		cfg.setProperty(MAX_ELAPSED_TIME, ""+maxElapsedTime);
	}
	
	// *** INNER METHODS *** //
	// ********************* //
	private String filterListToString(List<String> stateFilter) {
		String s = "[";
		for(int i = 0 ; i < stateFilter.size(); i++) {
			s += stateFilter.get(i);
			if(i < stateFilter.size()-1) s+=",";
		}
		return s + "]";
	}
}
