package ai.lazuli.cesga.data.collection;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;

import ai.lazuli.cesga.logging.Logger;

/**
 * Collects data after a concrete time interval has passed between each collection.
 * @author Alberto M. Esmorís Pena
 *
 */
public class Collector {
	// *** CONSTANTS *** //
	// ***************** //
	private final static int DEFAULT_COLLECTION_COMMAND_EXPECTED_EXIT_STATUS = 0;
	
	// *** ATTRIBUTES *** //
	// ****************** //
	private long interval; // Time interval in seconds
	private Long lastCollectionTime; // Time (EPOCH format) when the last happened
	private boolean collecting;
	private String collectionCommand[];
	private int collectionCommandExpectedExitStatus;
	private Thread thread = null;
	private AfterCollectionStrategy acs;
	
	// *** CONSTRUCTOR *** //
	// ******************* //
	/**
	 * @param interval Time interval between data collections.
	 * @param doFirstCollection When TRUE the first collection will be performed
	 * (even if collecting is FALSE) just after the Collector instance has been built.
	 * @param collecting When TRUE this collector will get into collecting mode (which means
	 * after time interval has elapsed after the lastCollectionTime a collection will be done),
	 * when FALSE this collector will be inactive.
	 * When FALSE the first collection will be done after interval has elapsed. 
	 */
	public Collector(long interval, boolean doFirstCollection, boolean collecting, String collectionScriptPath,
			String collectionOutputPath, AfterCollectionStrategy acs)
			throws FileNotFoundException, IOException, InterruptedException, CollectionException {
		setInterval(interval);
		setLastCollectionTime(System.currentTimeMillis()/1000);
		setCollecting(collecting);
		setCollectionCommand(new String[]{"bash", collectionScriptPath, collectionOutputPath});
		setCollectionCommandExpectedExitStatus(DEFAULT_COLLECTION_COMMAND_EXPECTED_EXIT_STATUS);
		setAfterCollectionStrategy(acs);
		
		if(doFirstCollection) {
			collect();
		}
	}

	// *** GETTERS and SETTERs *** //
	// *************************** //
	public long getInterval() {
		return interval;
	}

	public void setInterval(long interval) {
		this.interval = interval;
	}

	public Long getLastCollectionTime() {
		return lastCollectionTime;
	}

	public void setLastCollectionTime(Long lastCollectionTime) {
		this.lastCollectionTime = lastCollectionTime;
	}

	public boolean isCollecting() {
		return collecting;
	}

	public void setCollecting(boolean collecting) {
		this.collecting = collecting;
	}
	
	public String getCollectionScriptPath() {
		return getCollectionCommand()[1];
	}

	public void setCollectionScriptPath(String collectionScriptPath) {
		getCollectionCommand()[1] = collectionScriptPath;
	}
	
	public String getCollectionOutputPath() {
		return getCollectionCommand()[2];
	}
	
	public void setCollectionOutputPath(String collectionOutputPath) {
		getCollectionCommand()[2] = collectionOutputPath;
	}
	
	public String[] getCollectionCommand() {
		return collectionCommand;
	}

	public void setCollectionCommand(String collectionCommand[]) {
		this.collectionCommand = collectionCommand;
	}
	
	public int getCollectionCommandExpectedExitStatus() {
		return collectionCommandExpectedExitStatus;
	}

	public void setCollectionCommandExpectedExitStatus(int collectionCommandExpectedExitStatus) {
		this.collectionCommandExpectedExitStatus = collectionCommandExpectedExitStatus;
	}
	
	public AfterCollectionStrategy getAfterCollectionStrategy() {
		return acs;
	}
	
	public void setAfterCollectionStrategy(AfterCollectionStrategy acs) {
		this.acs = acs;
	}
	
	// *** METHODS *** //
	// *************** //
	public void collect() throws FileNotFoundException, IOException,
			InterruptedException, CollectionException {
		File collectionScript = new File(getCollectionScriptPath());
		if(!collectionScript.exists()) {
			throw new FileNotFoundException(String.format(
					"Collection script file '%s' could NOT be found.",
					collectionScript.getAbsolutePath()));
		}
		
		if(!collectionScript.canRead() || !collectionScript.canExecute()) {
			throw new SecurityException(String.format(
					"Collection script file '%s' lacks from READ or EXECUTE permissions.",
					collectionScript.getAbsolutePath()));
		}
		
		Process p;
		try {
			p = Runtime.getRuntime().exec(getCollectionCommand());
		}catch(Exception ex) {
			String cmd = "";
			for(int i = 0 ; i < getCollectionCommand().length ; i++) {
				cmd += getCollectionCommand()[i];
			}
			Logger.printf("Could not exec collection command: '%s'\n", cmd);
			throw ex;
		}
		int code = p.waitFor();
		
		if(code != getCollectionCommandExpectedExitStatus()) {
			// Collection FAILED
			File f = new File("");
			String command = "";
			for(String str : getCollectionCommand())
				command += str+" ";
			BufferedReader in = new BufferedReader(new InputStreamReader(p.getInputStream()));
			String msg = "";
			String read;
			while((read=in.readLine()) != null)
				msg += read;
			in.close();
			Logger.printf("Failed to colect, with exit code %d\nWorking directory: \"%s\""+
					"\nCollection command: %s\nOutput: \"%s\"\n",
					code,f.getAbsolutePath(),command,msg);
			throw new CollectionException(String.format("Failed to colect, with exit code %d",
					code));
		}
		else {
			// Collection SUCCESS
			setLastCollectionTime(System.currentTimeMillis()/1000);
			BufferedReader in = new BufferedReader(new InputStreamReader(p.getInputStream()));
			String read, csvPath = null;
			while ((read = in.readLine()) != null) {
				csvPath = read;
			}
			in.close();
			Logger.printf("(%d) Data successfully collected to \"%s\"\n",
					System.currentTimeMillis()/1000,
					csvPath);
			
			acs.apply(csvPath);
		}
		
	}
	
	/**
	 * Starts a thread which will perform a collect operation when specified time interval
	 * has elapsed since the last collection.
	 * NOTICE that started thread will only collect data when isCollecting returns TRUE.
	 * ALSO when isCollecting returns FALSE the thread will eventually end.
	 */
	public void start() {
		if(thread==null) {
			thread = new CollectorThread(this);
		}
		
		thread.start();
		
	}
}
