package ai.lazuli.cesga.data.collection;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Hashtable;
import java.util.List;

import ai.lazuli.cesga.Config;
import ai.lazuli.cesga.data.ModelSynchronizer;
import ai.lazuli.cesga.data.Notifier;
import ai.lazuli.cesga.logging.Logger;
import ai.lazuli.utils.classificationrules.Classification;
import ai.lazuli.utils.classificationrules.ClassificationModel;
import ai.lazuli.utils.classificationrules.ClassificationModelParser;
import ai.lazuli.utils.classificationrules.ClassificationModelParserException;
import ai.lazuli.utils.classificationrules.ClassificationRuleException;
import ai.lazuli.utils.csv.CSV;
import ai.lazuli.utils.csv.CSVField;
import ai.lazuli.utils.csv.CSVIO;

public class PredictAfterCollection implements AfterCollectionStrategy {
	// *** CONSTANTS *** //
	// ***************** //
	public final static String FIELD_ID_JOB = "id_job";
	public final static String FIELD_ACC = "acc";
	public final static String FIELD_PREDICTED_CLASS = "predicted_class";
	
	// *** ATTRIBUTES *** //
	// ****************** //
	private String shortModelPath; // Path to short model rules file to use
	private ModelSynchronizer shortSync; // Must correspond with the short model
	private String longModelPath; // Path to long model rules file to use
	private ModelSynchronizer longSync; // Must correspond with the long model
	private String predictionDir; // Directory to export generated predictions inside
	private Collector predictor;
	private Config cfg;
	private Notifier notifier;
	
	// *** CONSTRUCTOR *** //
	// ******************* //
	public PredictAfterCollection(String shortModelPath, String longModelPath, ModelSynchronizer shortSync,
			ModelSynchronizer longSync, String predictionDir, Collector predictor, Config cfg) {
		setShortModelPath(shortModelPath);
		setLongModelPath(longModelPath);
		setShortSync(shortSync);
		setLongSync(longSync);
		setPredictionDir(predictionDir);
		setPredictor(predictor);
		setConfig(cfg);
		setNotifier(Notifier.getInstance(getConfig()));
	}
	
	// *** GETTERs and SETTERs *** //
	// *************************** //
	public String getShortModelPath() {
		return shortModelPath;
	}


	public void setShortModelPath(String shortModelPath) {
		this.shortModelPath = shortModelPath;
	}


	public String getLongModelPath() {
		return longModelPath;
	}

	public void setLongModelPath(String longModelPath) {
		this.longModelPath = longModelPath;
	}

	public ModelSynchronizer getShortSync() {
		return shortSync;
	}

	public void setShortSync(ModelSynchronizer shortSync) {
		this.shortSync = shortSync;
	}
	
	public ModelSynchronizer getLongSync() {
		return longSync;
	}

	public void setLongSync(ModelSynchronizer longSync) {
		this.longSync = longSync;
	}

	public String getPredictionDir() {
		return predictionDir;
	}


	public void setPredictionDir(String predictionDir) {
		this.predictionDir = predictionDir;
	}
	
	public Collector getPredictor() {
		return predictor;
	}

	public void setPredictor(Collector predictor) {
		this.predictor = predictor;
	}

	public Config getConfig() {
		return cfg;
	}

	public void setConfig(Config cfg) {
		this.cfg = cfg;
	}

	public Notifier getNotifier() {
		return notifier;
	}

	public void setNotifier(Notifier notifier) {
		this.notifier = notifier;
	}

	// *** APPLY *** //
	// ************* //
	@Override
	public void apply(String csvPath) {
		try {
			ClassificationModel scm = obtainShortClassificationModel();
			ClassificationModel lcm = obtainLongClassificationModel();
			CSV input = CSVIO.read(csvPath, true, null);
			List<CSVField> outputHeader = new ArrayList<>(0);
			outputHeader.add(new CSVField(CSVField.TYPE_STRING, FIELD_ID_JOB));
			outputHeader.add(new CSVField(CSVField.TYPE_STRING, FIELD_PREDICTED_CLASS));
			outputHeader.add(new CSVField(CSVField.TYPE_DECIMAL, FIELD_ACC));
			CSV output = new CSV(outputHeader);
			List<CSVField> notifierHeader = new ArrayList<>(input.getHeader());
			notifierHeader.add(new CSVField(CSVField.TYPE_DECIMAL, FIELD_ACC));
			notifierHeader.add(new CSVField(CSVField.TYPE_STRING, FIELD_PREDICTED_CLASS));
			CSV notifierRecords = new CSV(notifierHeader);
			for(int i = 0 ; i < input.size() ; i++) {
				Hashtable<String, Object> inputRecord = input.getRecordAsHashtable(i);
				Classification cl = null;
				try {
					cl = scm.classify(inputRecord);	
				}catch(ClassificationRuleException crex) {
					try {
						cl = lcm.classify(inputRecord);
					}catch(ClassificationRuleException crex2) {
						continue;
					}
				}
				List<Object> notifierRecord = input.getRecord(i);
				notifierRecord.add(cl.getAccuracy());
				notifierRecord.add(cl.getClassName());
				notifierRecords.addRecord(notifierRecord);
				
				List<Object> outputRecord = new ArrayList<>(0);
				outputRecord.add(inputRecord.get(FIELD_ID_JOB));
				outputRecord.add(cl.getClassName());
				outputRecord.add(cl.getAccuracy());
				output.addRecord(outputRecord);
			}
			Calendar cal = Calendar.getInstance();
			String predictionPath = String.format("%s%sPREDICTIONS_%d-%d-%dT%d:%d:%d.csv",
					getPredictionDir(),
					File.separator,
					cal.get(Calendar.YEAR),
					cal.get(Calendar.MONTH)+1,
					cal.get(Calendar.DAY_OF_MONTH),
					cal.get(Calendar.HOUR),
					cal.get(Calendar.MINUTE),
					cal.get(Calendar.SECOND));
			CSVIO.write(output, predictionPath, false);
			updateLastCollectionEpoch();
			notifier.notify(notifierRecords);
		}catch(Exception ex) {
			Logger.printException(ex);
		}
	}
	
	// *** INNER METHODS *** //
	// ********************* //
	private void updateLastCollectionEpoch() {
		String oldCmd[] = getPredictor().getCollectionCommand();
		String newCmd[] = Arrays.copyOf(oldCmd, oldCmd.length);
		newCmd[newCmd.length-3] = oldCmd[oldCmd.length-2]; // $2
		newCmd[newCmd.length-2] = (System.currentTimeMillis()/1000)+""; // $3
		cfg.setPredictorLastCollectionEpoch(Long.parseLong(newCmd[newCmd.length-2]));
		cfg.save();
		getPredictor().setCollectionCommand(newCmd);
	}
	
	private ClassificationModel obtainShortClassificationModel() throws ClassificationModelParserException{
		while(!getShortSync().tryModelReadOperation()) {
			try {
				Thread.sleep(5000);
			}catch(InterruptedException iex) {
				Logger.printException(iex);
			}
		}
		ClassificationModel scm = ClassificationModelParser.parseFromFile(getShortModelPath());
		getShortSync().releaseModelReadOperation();
		return scm;
	}
	
	private ClassificationModel obtainLongClassificationModel() throws ClassificationModelParserException{
		while(!getLongSync().tryModelReadOperation()) {
			try {
				Thread.sleep(5000);
			}catch(InterruptedException iex) {
				Logger.printException(iex);
			}
		}
		ClassificationModel lcm = ClassificationModelParser.parseFromFile(getLongModelPath());
		getLongSync().releaseModelReadOperation();
		return lcm;
	}

}
