package ai.lazuli.utils.classificationrules;

import java.io.DataInputStream;
import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import ai.lazuli.utils.classificationrules.ClassificationRuleCondition.LogicalOperator;

public class ClassificationModelParser {
	// *** METHODS *** //
	// *************** //
	
	/**
	 * Parses classification rules content (can be obtained from a file, written in the code itself, etc.)
	 * Syntax for this content is as follows:<br>
	 * 	<b>- Lines which begin with '#' are comments</b><br>
	 * 	<b>- Any other content must be a rule which respects the following syntax:</b><br>
	 * 		Rule &lt;String ID&gt;:	(&lt;DOUBLE&gt;, lift &lt;DOUBLE LIFT&gt;)<br>
	 * 		&lt;CONDITION 1&gt;<br>
	 * 		...<br>
	 * 		&lt;CONDITION N&gt;<br>
	 * 		-> class &lt;CLASS NAME&gt; [&lt;DOUBLE ACC&gt;]
	 *	<b>- Conditions syntax is as follows:</b><br>
	 *		&lt;STRING FIELD NAME&gt; &lt;LOGICAL OPERATOR&gt; &lt;COMPARISSON VALUE&gt;
	 *	<b>- Default classification rule syntax is as follows:<br>
	 *	 	Default class: &lt;CLASS NAME&gt;
	 */
	public static ClassificationModel parse(String content) throws ClassificationModelParserException {
		String lines[] = content.replace("\r","").split("\n");
		
		String modelId = null;
		List<ClassificationRules> rulesets = new ArrayList<>(0);
		int rulesetId = 0; // Incremental ruleset numerical ID component
		int ruleId = 0; // Incremental rule numerical ID component
		
		// Ruleset being parsed (when null, no ruleset is being parsed)
		ClassificationRules ruleset = null;
		// Default classification for the ruleset being parsed
		Classification dc = null;
		
		// Rule being parsed (when null, no rule is being parsed)
		ClassificationRule rule = null;
		// Conditions being parsed (when null, no conditions are being parsed)
		List<ClassificationRuleCondition> conds = null;
		
		// Condition being parsed (when null, no condition is being parsed)
		ClassificationRuleCondition cond = null;
		
		for(String line : lines) {
			line = line.trim();
			while(line.contains("  ")) line = line.replaceAll("  ", " ");
			Character fc = null; // First Character
			for(char c : line.toCharArray()) {
				if(c != '\t' && c != '\n' && c != '\r') {
					fc = c;
					break;
				}
			}
			if (fc == null || ((int)fc) == 0 || fc == '#') continue;
			
			String split[] = line.split("=");
			String first = split[0];
			
			if(first.equalsIgnoreCase("id")) {
				// Parse model ID
				String val = split[1].substring(
						split[1].indexOf("\"")+1,
						split[1].lastIndexOf("\"")
				);
				if(modelId!=null) {
					throw new ClassificationModelParserException(String.format(
							"Rules ID '%s' has been parsed but rules id '%s' was parsed before.\n",
							val, modelId));
				}
				modelId = val;
			}
			else if(first.equalsIgnoreCase("rules")) {
				// ------------------------------------------------------ //
				//  *****  Parse rule set START (rules + default)  *****  //
				// ------------------------------------------------------ //
				rulesetId++;
				boolean insideString = false;
				boolean escapingChar = false;
				boolean parsingFieldName = true;
				String field = "";
				String val = "";
				for(int i = 0 ; i < line.length() ; i++) {
					// BEGIN
					char c = line.charAt(i);
					if(c=='"' && !escapingChar) {
						insideString = !insideString;
						if(!insideString) {
							// String (field value) ended
							if(field.equals("rules")) {
								ruleset = new ClassificationRules();
								ruleset.setId(String.format("RULES %d",rulesetId));
							}
							else if(field.equals("default")) {
								dc = new Classification(val, 0.0);
								ruleset.setDefaultClassification(dc);
								rulesets.add(ruleset);
								break;
							}
							
							field = "";
							val = "";
							parsingFieldName = true;
						}
						continue;
					}
					
					// MIDDLE
					if(parsingFieldName) {
						// Parsing field name
						if(c==' ' || c=='=') continue;
						field += c;
						if(field.equals("rules") ||
								field.equals("default")) {
							parsingFieldName = false;
							continue;
						}
					}
					else if (insideString) {
						// Parsing field value
						val += c;
					}
					
					
					// END
					escapingChar = false;
					if(insideString && !escapingChar && c=='\\')
						escapingChar = true;
				}
				
			}
			else if(first.equalsIgnoreCase("conds")) {
				if(conds!=null) {
					// Add previous conds to previous rule before parsing next rule
					rule.setConditions(conds);
				}
				
				ruleId++;
				// ----------------------------------------------------------------------------- //
				//  *****  Parse rule set DEFINITION (conds + cover + ok + lift + class)  *****  //
				// ----------------------------------------------------------------------------- //
				boolean insideString = false;
				boolean escapingChar = false;
				boolean parsingFieldName = true;
				String field = "";
				String val = "";
				
				double cover = -1.0; // Number of cases covered by the rule
				double ok = -1.0; // Number of ok classified cases by the rule
				double lift = -1.0; // Estimated accuracy considering probability of application for the rule
				for(int i = 0 ; i < line.length() ; i++) {
					// BEGIN
					char c = line.charAt(i);
					if(c=='"' && !escapingChar) {
						insideString = !insideString;
						if(!insideString) {
							// String (field value) ended
							if(field.equals("conds")) {
								conds = new ArrayList<>(Integer.parseInt(val));
							}
							else if(field.equals("cover")) {
								cover = Double.parseDouble(val);
							}
							else if(field.equals("ok")) {
								ok = Double.parseDouble(val);
							}
							else if(field.equals("lift")) {
								lift = Double.parseDouble(val);
							}
							else if(field.equals("class")) {
								if(cover < .0)
									throw new ClassificationModelParserException(
											"No 'cover' field was found when parsing rule definition");
								if(ok < .0)
									throw new ClassificationModelParserException(
											"No 'ok' field was found when parsing rule definition");
								if(lift < .0)
									throw new ClassificationModelParserException(
											"No 'lift' field was found when parsing rule definition");
								double err = cover-ok;
								double acc = (cover - err + 1.0) / (cover+2.0);
								Classification cl = new Classification(val, acc);
								String id = String.format("RULE %d-%d", rulesetId, ruleId);
								rule = new ClassificationRule(id, (int)cover, (int)ok, lift, conds, cl);
								ruleset.addRule(rule);
								break;
							}
							
							field = "";
							val = "";
							parsingFieldName = true;
						}
						continue;
					}
					
					// MIDDLE
					if(parsingFieldName) {
						// Parsing field name
						if(c==' ' || c=='=') continue;
						field += c;
						if(field.equals("conds") ||
								field.equals("cover") ||
								field.equals("ok") ||
								field.equals("lift") ||
								field.equals("class")) {
							parsingFieldName = false;
							continue;
						}
					}
					else if (insideString) {
						// Parsing field value
						val += c;
					}
					
					
					// END
					escapingChar = false;
					if(insideString && !escapingChar && c=='\\')
						escapingChar = true;
				}
			}
			else if(first.equalsIgnoreCase("type")) {
				// -------------------------------------------------------------------- //
				//  *****  Parse rule set CONDITION (type + att + cut + result)  *****  //
				// -------------------------------------------------------------------- //
				boolean insideString = false;
				boolean escapingChar = false;
				boolean parsingFieldName = true;
				String field = "";
				String val = "";
				String valList[] = null; // Alternative to val used to parse elts
				
				// Attribute implied in the condition
				String att = null;
				
				// Numerical cut for the condition (exclusive respect 'str' and 'elts')
				Double cut = null;
				// When cut is specified, a logical operator must be specified too
				String lop = null;
				
				// String used to compare by the condition (exclusive respect 'cut' and 'elts')
				String str = null;
				
				/* List of strings used to compare by the condition
					(exclusive respect 'cut' and 'str') */
				List<String> elts = null;
				for(int i = 0 ; i < line.length() ; i++) {
					// BEGIN
					char c = line.charAt(i);
					if(c=='"' && !escapingChar) {
						insideString = !insideString;
						if(!insideString) {
							// String (field value) ended
							if(field.equals("type")) {
								// Do nothing
							}
							else if(field.equals("att")) {
								att = val;
							}
							else if(field.equals("cut")) {
								cut = Double.parseDouble(val);
								cond = new ClassificationRuleCondition(att,
										ClassificationRuleCondition.LogicalOperator.eq,
										ClassificationRuleCondition.ValueType.number,
										cut);
								// Breaks when field.equals("result")
							}
							else if(field.equals("val")) {
								str = val;
								cond = new ClassificationRuleCondition(att,
										ClassificationRuleCondition.LogicalOperator.eq,
										ClassificationRuleCondition.ValueType.string,
										str);
								conds.add(cond);
								break;
							}
							else if(field.equals("result")) {
								lop = val;
								if(lop.equals("<"))
									// < ~ <= because of the 'cut' mechanic
									cond.setLop(LogicalOperator.lte);
								else if(lop.equals(">"))
									cond.setLop(LogicalOperator.gt);
								else
									throw new ClassificationModelParserException(String.format(
											"Unexpected logical operator '%s' when parsing cut-result condition.\n"+
											"\tExpected operators are: {'<' , '>'}"));
								conds.add(cond);
								break;
							}
							
							
							field = "";
							val = "";
							parsingFieldName = true;
						}
						continue;
					}
					
					// MIDDLE
					if(parsingFieldName) {
						// Parsing field name
						if(c==' ' || c=='=') continue;
						field += c;
						
						if(field.equals("elts")) {
							// When field is elts a different behavior is needed
							String substr = line.substring(i+2);
							valList = substr.split(",");
							for(int j = 0 ; j < valList.length ; j++)
								valList[j] = valList[j].replace("\"","");
							
							elts = new ArrayList<>(Arrays.asList(valList));
							cond = new ClassificationRuleCondition(att,
									ClassificationRuleCondition.LogicalOperator.in,
									ClassificationRuleCondition.ValueType.list,
									elts);
							conds.add(cond);
							break;
						}
						else if(field.equals("type") ||
								field.equals("att") ||
								field.equals("cut") ||
								field.equals("val") ||
								field.equals("elts") ||
								field.equals("result")) {
							parsingFieldName = false;
							continue;
						}
					}
					else if (insideString) {
						// Parsing field value
						val += c;
					}
					
					
					// END
					escapingChar = false;
					if(insideString && !escapingChar && c=='\\')
						escapingChar = true;
				}
			}
			
			
			
		}
		
		if(rulesets.size()>0) return new ClassificationModel(modelId, rulesets);
		
		throw new ClassificationModelParserException("Classification rules could not be parsed");
	}
	
	public static ClassificationModel parseFromFile(String path) throws ClassificationModelParserException {
		try {
			String content = "";
			DataInputStream in = new DataInputStream(new FileInputStream(path));
			
			byte buffer[] = new byte[4096];
			while(in.read(buffer)!=-1) {
				content += new String(buffer);
				buffer = new byte[4096];
			}
			
			in.close();
			
			return parse(content);
		}catch(Exception ex) {
			throw new ClassificationModelParserException(String.format(
					"Could not parse classification rules from file: '%s'",path),ex);
					
		}
	}
	
}
