package ai.lazuli.utils.classificationrules;

import java.util.Hashtable;
import java.util.List;
import java.util.Locale;

public class ClassificationRuleCondition {
	public enum LogicalOperator{
		gt, // Greater than (value is number)
		gte, // Greater than or equal to (value is number)
		lt, // Less than (value is number)
		lte, // Less than or equal to (value is number)
		eq, // Equals (value may be for string or number)
		in // Field is contained in list (value is list of strings)
	}
	
	public enum ValueType{
		number, // Value is Double
		string, // Value is String
		list // Value is List<String>
	}
	
	// *** ATTRIBUTES *** //
	// ****************** //
	private String field;
	private LogicalOperator lop;
	private ValueType vt;
	private Object value;

	// *** CONSTRUCTORS *** //
	// ******************** //
	public ClassificationRuleCondition(String field, LogicalOperator lop, ValueType vt, Object value) {
		super();
		setField(field);
		setLop(lop);
		setVt(vt);
		setValue(value);
	}
	
	// *** GETTERs and SETTERs *** //
	// *************************** //
	public String getField() {
		return field;
	}
	
	public void setField(String field) {
		this.field = field;
	}

	public LogicalOperator getLop() {
		return lop;
	}
	
	public String getLopAsString() {
		switch(getLop()) {
		case eq:
			return "=";
		case gt:
			return ">";
		case gte:
			return ">=";
		case in:
			return "in";
		case lt:
			return "<";
		case lte:
			return "<=";
		default:
			return "????";
		}
	}

	public void setLop(LogicalOperator lop) {
		this.lop = lop;
	}

	public ValueType getVt() {
		return vt;
	}

	public void setVt(ValueType vt) {
		this.vt = vt;
	}

	public Object getValue() {
		return value;
	}

	public void setValue(Object value) {
		this.value = value;
	}
	
	// *** TO STRING *** //
	// ***************** //
	@Override
	public String toString() {	
		if(getVt()==ValueType.number) 
			return String.format(Locale.ROOT, "\t%s %s %f\n", getField(),
					getLopAsString(), (Double) getValue());
		else if(getVt()==ValueType.list) {
			String listStr = "{";
			@SuppressWarnings("unchecked")
			List<String> list = (List<String>) getValue();
			for (int i = 0 ; i < list.size() ; i++) {
				listStr += list.get(i);
				if(i < list.size()-1) listStr += ", ";
			}
			listStr +="}";

			return String.format(Locale.ROOT, "\t%s %s %s\n", getField(),
					getLopAsString(), listStr);
		}
		else
			return String.format(Locale.ROOT, "\t%s %s %s\n", getField(),
					getLopAsString(), getValue());
	}

	// *** METHODS *** //
	// *************** //
	/**
	 * @return TRUE if this condition is satisfied by the attributes in the received hashtable.
	 * FALSE otherwise.
	 */
	public boolean isSatisfiedBy(Hashtable<String, Object> set) throws ClassificationRuleException {
		switch(getVt()) {
			case list:
				if(getLop()!=LogicalOperator.in)
					throw new ClassificationRuleException(String.format(
							"Unexpected logical operator:\n"+
							"\tExpected 'in' but has '%s' instead\n", getLopAsString()));
				String receivedValue = (String) set.get(getField());
				@SuppressWarnings("unchecked")
				List<String> thisList = (List<String>) getValue();
				if(thisList.contains(receivedValue))
					return true;
				break;
			case number:
				Double receivedNumber = Double.parseDouble(set.get(getField()).toString());
				Double thisNumber = (Double) getValue();
				switch(getLop()) {
					case eq:
						return receivedNumber.compareTo(thisNumber) == 0;
					case gt:
						return receivedNumber.compareTo(thisNumber) > 0;
					case gte:
						return receivedNumber.compareTo(thisNumber) >= 0;
					case lt:
						return receivedNumber.compareTo(thisNumber) < 0;
					case lte:
						return receivedNumber.compareTo(thisNumber) <= 0;
					case in:
					default:
						throw new ClassificationRuleException(String.format(
								"Unexpected logical operator:\n"+
								"\tExpected 'in' but has '%s' instead\n", getLopAsString()));
				}
			case string:
				if(getLop()!=LogicalOperator.eq)
					throw new ClassificationRuleException(String.format(
							"Unexpected logical operator:\n"+
							"\tExpected '=' but has '%s' instead\n", getLopAsString()));
					String receivedString = (String) set.get(getField());
					String thisString = (String) getValue();
					return receivedString.equals(thisString);
			default:
				throw new ClassificationRuleException(String.format(
						"Unexpected condition:\n\t%s\n",
						toString()));
		}
		
		return false;
	}
	
	
}
