package ai.lazuli.utils.classificationrules;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;

public class ClassificationRules {
	// *** ATTRIBUTES *** //
	// ****************** //
	private String id;
	private List<ClassificationRule> rules;
	private Classification defaultClassification;
	
	// *** CONSTRUCTOR *** //
	// ******************* //
	public ClassificationRules(String id, List<ClassificationRule> rules,
			Classification defaultClassification) {
		this.id = id;
		this.rules = rules;
		this.defaultClassification = defaultClassification;
	}
	
	public ClassificationRules() {
		this.id = null;
		this.rules = new ArrayList<>(0);
		this.defaultClassification = null;
	}
	
	// *** GETTERs and SETTERs *** //
	// *************************** //
	public String getId() {
		return id;
	}

	public void setId(String id) {
		this.id = id;
	}
	
	public ClassificationRule getRule(int index) {
		return rules.get(index);
	}
	
	public ClassificationRule getRuleById(String id) {
		for(ClassificationRule r : rules)
			if(r.getId().equals(id)) return r;
		return null;
	}
	
	public void setRule(int index, ClassificationRule rule) {
		rules.set(index,  rule);
	}
	
	public List<ClassificationRule> getRules() {
		return rules;
	}
	
	public void setRules(List<ClassificationRule> rules) {
		this.rules = rules;
	}
	
	public Classification getDefaultClassification() {
		return defaultClassification;
	}

	public void setDefaultClassification(Classification defaultClassification) {
		this.defaultClassification = defaultClassification;
	}

	// *** METHODS *** //
	// *************** //
	public void addRule(ClassificationRule rule) {
		rules.add(rule);
	}
	
	public void delRule(int index) {
		rules.remove(index);
	}
	
	public void delRule(ClassificationRule rule) {
		rules.remove(rule);
	}
	
	public int size() {
		return rules.size();
	}
	
	/**
	 * Applies the classification rules to the received data set.
	 * @return Final classification object, or null if no classification could be done.
	 */
	public Classification classify(Hashtable<String, Object> set) throws ClassificationRuleException{
		/*
		 * Key (String): Classification class name
		 * Value (Double): Aggregated average accuracy from all applicable rules 
		 */
		Hashtable<String, Double> accs = new Hashtable<>();
		Hashtable<String, Double> n = new Hashtable<>(); 
		
		for(ClassificationRule cr : rules) {
			if(cr.isSatisfiedBy(set)) {
				Classification cls = cr.getClassification();
				String className = cls.getClassName();
				Double acc = cls.getAccuracy();
				if(accs.containsKey(className)) {
					n.put(className, n.get(className)+1.0);
					acc += accs.get(className);
					accs.put(className,acc);
				}
				else {
					accs.put(className, acc);
					n.put(className, 1.0);
				}
			}
		}
		
		if(accs.size()<1) return defaultClassification;
		
		Enumeration<String> keys = accs.keys();
		String className = keys.nextElement();
		Double accuracy = accs.get(className);
		
		while(keys.hasMoreElements()) {
			String candidateClassName = keys.nextElement();
			Double candidateAccuracy = accs.get(candidateClassName);
			if(candidateAccuracy > accuracy) {
				className = candidateClassName;
				accuracy = candidateAccuracy; 
			}
		}
		
		
		return new Classification(className, accuracy/n.get(className));
	}
	
	// *** toString *** //
	// **************** //
	@Override
	public String toString() {
		String str = String.format("Default class: %s\n",
			getDefaultClassification().getClassName());
		for(ClassificationRule rule : rules)
			str += rule.toString();
		return str;
	}
	
}
