package ai.lazuli.utils.csv;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;


public class CSV {
	// *** ATTRIBUTES *** //
	// ****************** //
	private List<CSVField> header;
	private List<List<Object>> records = null;
	
	// *** CONSTRUCTORS *** //
	// ******************** //
	public CSV(List<CSVField> header) throws InvalidCSVRecordException{
		this.header = header;
	}
	
	public CSV(List<CSVField> header, List<List<Object>> records) throws InvalidCSVRecordException{
		this.header = header;
		addRecords(records);
	}
	
	// *** GETTERs and SETTERs *** //
	// *************************** //
	public List<CSVField> getHeader(){
		return header;
	}
	
	public List<List<Object>> getRecords(){
		return records;
	}
	
	public List<Object> getRecord(int index){
		return records.get(index);
	}
	
	/**
	 * Obtains the record at the specified index as a hashtable where
	 * header field names are the key and its corresponding value the object.
	 */
	public Hashtable<String, Object> getRecordAsHashtable(int index){
		Hashtable<String, Object> ht = new Hashtable<>();
		for(int i = 0 ; i < getHeader().size() ; i++)
			ht.put(getHeader().get(i).getName(), getRecord(index).get(i));
		return ht;
	}
	
	public void setRecord(int index, List<Object> record) {
		records.set(index, record);
	}
	
	/**
	 * @return The object itself.
	 * NULL if it was NaN.
	 */
	public Object getValue(int record, int field) {
		return records.get(record).get(field);
	}
	
	public Object getValue(int record, String fieldName) {
		for(int i = 0 ; i < header.size() ; i++) {
			String name = header.get(i).getName();
			if(name.equals(fieldName))
				return getValue(record, i);
		}
		
		throw new NullPointerException(String.format(
			"No field with name '%s' was found.", fieldName));
	}
	
	public String getValueAsString(int record, int field) {
		return (String) getValue(record, field);
	}
	public String getValueAsString(int record, String fieldName) {
		return (String) getValue(record, fieldName);
	}
	
	public Integer getValueAsInteger(int record, int field) {
		return (Integer) getValue(record, field);
	}
	public Integer getValueAsInteger(int record, String fieldName) {
		return (Integer) getValue(record, fieldName);
	}
	
	public Long getValueAsLong(int record, int field) {
		return (Long) getValue(record, field);
	}
	public Long getValueAsLong(int record, String fieldName) {
		return (Long) getValue(record, fieldName);
	}
	
	public Float getValueAsFloat(int record, int field) {
		return (Float) getValue(record, field);
	}
	public Float getValueAsFloat(int record, String fieldName) {
		return (Float) getValue(record, fieldName);
	}
	
	public Double getValueAsDouble(int record, int field) {
		return (Double) getValue(record, field);
	}
	public Double getValueAsDouble(int record, String fieldName) {
		return (Double) getValue(record, fieldName);
	}
	
	public void setValue(int record, int field, Object value) {
		records.get(record).set(field, value);
	}
	
	public boolean isNaN(int record, int field) {
		return records.get(record).get(field) == null;
	}
	
	// *** METHODS *** //
	// *************** //
	public void addRecords(List<List<Object>> records) throws InvalidCSVRecordException{
		for(List<Object> record : records)
			addRecord(record);
	}
	
	public void addRecord(List<Object> record) throws InvalidCSVRecordException {
		validateRecord(record);
		if(records == null)
			records = new ArrayList<>(0);
		records.add(record);
	}
	
	public void delRecord(int index) {
		records.remove(index);
	}
	
	public void clearRecords(int index) {
		records.clear();
	}
	
	/**
	 * @return Number of records (-1 if records list was not even instantiated)
	 */
	public int size() {
		if(records == null) return -1;
		return records.size();
	}
	
	// *** TO STRING *** //
	// ***************** //
	public void print(PrintStream ps) {
		for(int i = 0 ; i < header.size() ; i++) {
			ps.print(header.get(i).getName());
			if(i < header.size()-1)
				ps.print(", ");
			else
				ps.println();
		}
		
		for(List<Object> record : records) {
			for(int i = 0 ; i < record.size(); i++) {
				ps.print(record.get(i));
				if(i < record.size() - 1)
					ps.print(", ");
			}
			ps.println();
		}
		
	}
	
	// *** INNER METHODS *** //
	// ********************* //
	/**
	 * Do nothing when the record is valid.
	 * Throw and InvalidCSVRecordException with an adequate exception message
	 * when the record was not valid.
	 */
	private void validateRecord(List<Object> record) throws InvalidCSVRecordException{
		if(record==null)
			throw new InvalidCSVRecordException("Record was null");
		else if(record.size() != header.size())
			throw new InvalidCSVRecordException(String.format(
					"Record size (%d) is different than header size (%d)",
					record.size(), header.size()));
		
		for(int i = 0 ; i < record.size() ; i++) {
			Object value = record.get(i);
			byte type = header.get(i).getType();
			
			// null is considered NaN, which may be acceptable or not, but it is valid anyway
			if (value == null) continue;
			
			switch(type) {
				case CSVField.TYPE_INTEGER:
					try {
						Long.parseLong(value.toString());
					}catch(NumberFormatException nfex) {
						throw new InvalidCSVRecordException(String.format(
								"Value '%s' was not an integer as expected",value));
					}
				case CSVField.TYPE_DECIMAL:
					try {
						Double.parseDouble(value.toString());
					}catch(NumberFormatException nfex) {
						throw new InvalidCSVRecordException(String.format(
								"Value '%s' was not a decimal as expected",value));
					}
					
			}
		}
	}
	
}

