package ai.lazuli.utils.csv;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

public class CSVIO {
	
	// *** METHODS *** //
	// *************** //
	/**
	 * Reads the CSV from the given path.
	 * @param headerFromFirstLine When TRUE, the header will be taken from the first line
	 * and header parameter will be ignored.
	 * @param header When headerFromFirstLine is FALSE, this parameter will be used as the
	 * header for the CSV file if different than null. If this is null and headerFromFirstLine
	 * is false, header will be built based on data suppositions made from first record parsing only.
	 */
	public static CSV read(String path, boolean headerFromFirstLine, List<CSVField> header) throws CSVIOException{
		File f = new File(path);
		if(!f.canRead())
			throw new CSVIOException(String.format("Can NOT read '%s'",path));
	
		
		CSV csv = null;
		List<CSVField> csvHeader = new ArrayList<CSVField>(0);
		List<List<Object>> csvRecords = new ArrayList<List<Object>>(0);
		
		
		BufferedReader in = null;
		try {
			in = new BufferedReader(new FileReader(f));
			String recordLine = null;
			
			// Header
			if(header != null) {
				csvHeader = header;
			}
			else {
				if(headerFromFirstLine) {
					String split[] = in.readLine().split(",");
					for(String name : split)
						csvHeader.add(new CSVField(CSVField.TYPE_STRING, name.replaceAll(" ", "")));
				}
				recordLine = in.readLine();
				String split[] = recordLine.split(",");
				if(split.length != csvHeader.size()) {
					throw new CSVIOException(String.format(
							"First record size (%d) and header size (%d) were not the same",
							split.length, csvHeader.size()));
				}
				for(int i = 0 ; i < split.length ; i++) {
					csvHeader.get(i).setType(
							determineFieldTypeByValue(split[i]));
				}
			}
			
			// Records
			if(recordLine == null) recordLine = in.readLine();
			if(recordLine == null)
				throw new CSVIOException(String.format(
						"Not even a single record was found when reading '%s'", path));
			do {
				if(recordLine.length()<1) continue; // skip empty lines
				ArrayList<Object> record = new ArrayList<>(0);
				String split[] = recordLine.split(",");
				if(split.length != csvHeader.size())
					throw new CSVIOException(String.format(
							"Record number of fields (%d) is not the same than header "
							+ "number of fields (%d)",
							split.length, csvHeader.size()));
				for(int i = 0 ; i < split.length ; i++) {
					byte type = csvHeader.get(i).getType();
					addValueToRecord(record, type, split[i]);
				}
				
				csvRecords.add(record);
			}while((recordLine=in.readLine())!=null);
			
			
			// CSV
			csv = new CSV(csvHeader, csvRecords);
		}catch(Exception ex) {
			throw new CSVIOException(ex);
		}finally{
			try {
				if(in!=null) in.close();
			}catch(Exception ex) {
				throw new CSVIOException(ex);
			}
		}
		
		
		return csv;
	}
	
	/**
	 * Writes the CSV to the given path.
	 * @param append if TRUE, writing operations will be done after the end of file.
	 * If FALSE, any data that existed in the specified path will be overwritten.
	 */
	public static void write(CSV csv, String path, boolean append) throws CSVIOException {
		PrintWriter out = null;
		try {
			File f = new File(path);
			if(!f.exists()) f.createNewFile();
			
			if (!f.canWrite())
				throw new CSVIOException(String.format("Can NOT write to path '%s'", path));
			
			
			out = new PrintWriter(new FileOutputStream(path, append));
			// Write header
			for(int i = 0 ; i < csv.getHeader().size() ; i++) {
				out.print(csv.getHeader().get(i).getName());
				if(i < csv.getHeader().size()-1) out.print(",");
			}
			out.println();
			
			// Write records
			for(int i = 0 ; i < csv.size() ; i++) {
				List<Object> record = csv.getRecord(i);
				for(int j = 0 ; j < record.size() ; j++) {
					// Write record
					Object v = record.get(j);
					out.print(v);
					if(j < record.size()-1) out.print(",");
				}
				out.println();
			}
			
		}catch(Exception ex) {
			throw new CSVIOException(ex);
		}finally{
			try {
				if(out!=null) out.close();
			}catch(Exception ex) {
				throw new CSVIOException(ex);
			}
		}
	}

	// *** INNER METHODS *** //
	// ********************* //
	/**
	 * Determines the field type analyzing the received value.
	 * @see ai.lazuli.utils.csv.CSVField
	 */
	private static byte determineFieldTypeByValue(String value) {
		try {
			Long.parseLong(value);
			return CSVField.TYPE_INTEGER;
		}catch(Exception ex) {}
		try {
			Double.parseDouble(value);
			return CSVField.TYPE_DECIMAL;
		}catch(Exception ex) {}
		
		return CSVField.TYPE_STRING;
	}
	
	/**
	 * Adds the parsed value to the record, considering it is of the given type.
	 */
	private static void addValueToRecord(List<Object>record, byte type, String value) {
		Object v = null;
		
		switch(type) {
			case CSVField.TYPE_INTEGER:
				try {
					v = Long.parseLong(value);
				}catch(Exception ex) {}
				break;
			case CSVField.TYPE_DECIMAL:
				try {
					v = Double.parseDouble(value);
				}catch(Exception ex) {}
				break;
			case CSVField.TYPE_STRING:
				v = value;
				break;
		}
		
		record.add(v);
	}
}

