package ai.lazuli.utils.files;

import java.io.File;
import java.io.IOException;
import java.nio.file.FileSystem;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.ArrayList;
import java.util.List;

public class FilePurger{
	// *** METHODS *** //
	// *************** //
	/**
	 * @see ai.lazuli.utils.files.FilePurger#purgeByCreationDate(File, long)
	 */
	public static List<File> purgeByCreationDate(String dir, long maxOld) throws IOException{
		return purgeByCreationDate(new File(dir), maxOld);
	}
	
	/**
	 * Purges files inside dir (including subdirectories) which were CREATED
	 * before [now - maxOld] (in seconds)
	 * @param dir Directory which contents will be purged when exceeding maxOld.
	 * @param maxOld Max old (in seconds) that files which will be purged must exceed.
	 * @return List of purged files
	 */
	public static List<File> purgeByCreationDate(File dir, long maxOld) throws IOException{
		List<File> pf = new ArrayList<>(0); // Purged Files
		
		
		FileSystem fs = FileSystems.getDefault(); // File System
		for(File f : dir.listFiles()) {
			Path path = fs.getPath(f.getAbsolutePath());
			BasicFileAttributes fa = Files.readAttributes(path,BasicFileAttributes.class);
			long now = System.currentTimeMillis()/1000;
			long creation = fa.creationTime().toMillis()/1000;
			if(now-creation > maxOld) {
				if(f.delete()) pf.add(f);
			}
		}
		
		return pf;
	}
	
	/**
	 * @see ai.lazuli.utils.files.FilePurger#purgeByModificationDate(File, long)
	 */
	public static List<File> purgeByModificationDate(String dir, long maxOld) throws IOException{
		return purgeByModificationDate(new File(dir), maxOld);
	}
	
	/**
	 * Works as purgeByCreationDate but considering last modification date instead of
	 * creation date.
	 * @see ai.lazuli.utils.files.FilePurger#purgeByCreationDate(File, long)
	 */
	public static List<File> purgeByModificationDate(File dir, long maxOld) throws IOException{
		List<File> pf = new ArrayList<>(0); // Purged Files
		
		
		FileSystem fs = FileSystems.getDefault(); // File System
		for(File f : dir.listFiles()) {
			Path path = fs.getPath(f.getAbsolutePath());
			BasicFileAttributes fa = Files.readAttributes(path,BasicFileAttributes.class);
			long now = System.currentTimeMillis()/1000;
			long modification = fa.lastModifiedTime().toMillis()/1000;
			if(f.isDirectory()) {
				pf.addAll(purgeByModificationDate(f, maxOld));
			}
			if(now-modification > maxOld) {
				if(f.delete()) pf.add(f);
			}
		}
		
		return pf;
	}
	
}
